#!/usr/bin/env python

"""
blackbox.py
A 'black-box' method for an analysis of Eos asteroid family.

"""

__author__ = "Miroslav Broz (miroslav.broz@email.cz)"
__version__ = "May 14th 2018" 

import sys
import math
import random
import subprocess


def pV_func(H, D):
    """Convert absolute magnitude and diameter to geometric albedo"""
    return 10.**(6.259 - 2.*math.log10(D) - 0.4*H)

def D_func(H, pV):
    """Convert absolute magnitude and geometric albedo to diameter"""
    return 10.**(0.5 * (6.259 - math.log10(pV) - 0.4*H))

def D_from_catalogue(l, pV):
    """Determine diameter from one catalogue line"""

    H_wise = l[45]
    pV_wise = l[49]
    D_wise = l[47]
    H_astorb = float(l[3])
    D_iras = l[6]

    if D_wise != "?":
        D = float(D_wise)
        pV = pV_wise
    elif D_iras != "?":
        D = float(D_iras)
        pV = pV_func(H_astorb,float(D_iras))
    else:
        try:
            pV = float(pV)
        except:
            if a_astdys != "?":
                a = float(a_astdys)
            else:
                a = float(a_astorb)
            if (a < 2.8):
                pV = 0.15
            else:
                pV = 0.05
        D = D_func(H_astorb,pV)
    return D


class Blackbox(object):
    """Blackbox to compare observed and synthetic data"""

    def __init__(self, file_observed = "Ktypes.dat", file_background = "background.dat",
        file_synthetic = "follow-sort-time.out", file_diameters = "yarko.in",
        seed = -3, pV = 0.158, factor = 2.0):
        """
        Prepare observed, background, and synthetic data

        file_observed   ... observed data (catalogue of Astorb + Astdys + Wise + Sloan)
        file_background ... background data (dtto)
        file_synthetic  ... synthetic family, output from Swift N-body integrator
        file_diameters  ... synthetic radii (R -> D), input for Swift
        seed   ... random seed
        pV     ... geometric albedo (if unknown)
        factor ... the ratio of total volume (Ktypes.dat) to background volume (background.dat)

        """

        self.file_observed   = file_observed
        self.file_background = file_background
        self.file_synthetic  = file_synthetic
        self.file_diameters  = file_diameters
        self.seed = seed
        self.pV = pV
        self.factor = factor

        random.seed(self.seed)

        self.da = None
        self.a1 = None
        self.a2 = None
        self.de = None
        self.e1 = None
        self.e2 = None
        self.di = None
        self.i1 = None
        self.i2 = None
        self.dD = None
        self.D1 = None
        self.D2 = None

        self.n1 = None
        self.n2 = None
        self.n3 = None
        self.n4 = None

        self.a_obs = None
        self.e_obs = None
        self.i_obs = None
        self.D_obs = None

        self.t = None
        self.a_syn = None
        self.e_syn = None
        self.i_syn = None
        self.D_syn = None

        self.a_scl = None
        self.e_scl = None
        self.i_scl = None
        self.D_scl = None

        self.D_bg = None

        self.bins = None
        self.list_of_synthetic = None
        self.list_of_background = None

        self.dN_obs = None
        self.dN_bg  = None
        self.dN_syn = None
        self.dN_scl = None

        self.boxes_obs = {}
        self.boxes_scl = {}


        self.define_boxes()
        self.read_observed()
        self.save_observed()
        self.observed_differential_sfd()
        self.count_boxes()
        self.save_boxes()
        self.read_background()
        self.background_differential_sfd()
        self.read_synthetic_diameters()
        self.read_synthetic_family()

    def analyse(self):
        """Analyse one segment of synthetic family data"""

        self.synthetic_differential_sfd()
        self.synthetic_background()
        self.rescale_synthetic()
        self.scaled_differential_sfd()
        self.save_scaled()
        self.count_boxes()
        self.save_boxes()
        self.create_images()
        chi2, n = self.chi2()
        return chi2, n

    def define_boxes(self):
        """Define boxes in (a_p, e_p, sin I_p) space"""

        a_J73 = 2.956
        a_J94 = 3.029

        self.da = (a_J94-a_J73)/3.
        self.a1 = a_J73-3.*self.da
        self.a2 = self.a1+15.*self.da
        self.de = 0.025
        self.e1 = 0.0
        self.e2 = 0.3+self.de
        self.di = 0.24
        self.i1 = 0.06
        self.i2 = 0.30
        self.dD = 120.
        self.D1 = 0.0
        self.D2 = 120.

        self.n1 = int((self.a2-self.a1)/self.da+0.5)
        self.n2 = int((self.e2-self.e1)/self.de+0.5)
        self.n3 = int((self.i2-self.i1)/self.di+0.5)
        self.n4 = int((self.D2-self.D1)/self.dD+0.5)
        n = self.n1*self.n2*self.n3*self.n4

        print "a1 = ", self.a1, " au"
        print "a2 = ", self.a2, " au"
        print "n1 = ", self.n1
        print "n2 = ", self.n2
        print "n3 = ", self.n3
        print "n4 = ", self.n4
        print "n = ", n

    def read_observed(self):
        """Read observed asteroids"""

        self.a_obs = []
        self.e_obs = []
        self.i_obs = []
        self.D_obs = []

        fin = open(self.file_observed, "r")

        for line in fin:
            l = line.split()
            self.a_obs.append(float(l[35]))
            self.e_obs.append(float(l[36]))
            self.i_obs.append(float(l[37]))
            self.D_obs.append(D_from_catalogue(l, self.pV))

        fin.close()
        #D_obs.sort()
        print "number_of_observed = " + str(len(self.a_obs))

    def save_observed(self):
        """Save observed data"""
        
        fout = file("obs.dat", "w")
        for i in xrange(0,len(self.a_obs)):
            fout.write("%8.3f %4d %.5f %.5f %.5f %.5f\n" % (0.0, i, self.a_obs[i], self.e_obs[i], self.i_obs[i], self.D_obs[i]))
        fout.close()

    def observed_differential_sfd(self):
        """Compute observed differential SFD"""

        D1_ = 0.
        D2_ = 20.
        dD_ = 1.

        self.bins = []
        self.dN_obs = []
        D = D1_
        eps = 1.e-6
        while D < D2_-eps:
            self.bins.append(D)
            self.dN_obs.append(0)
            D += dD_

        D1_ = 20.
        D2_ = 120.
        dD_ = 10.

        D = D1_
        while D < D2_-eps:
            self.bins.append(D)
            self.dN_obs.append(0)
            D += dD_

        for i in xrange(0,len(self.D_obs)):
            j = 1
            while self.D_obs[i] > self.bins[j]:
                j += 1
            self.dN_obs[j-1] += 1

        print "bins = " + str(self.bins)
        print "dN_obs = " + str(self.dN_obs)

    def count_boxes(self):
        """Count numbers of asteroids in boxes"""

        if len(self.boxes_obs) == 0:
            a = self.a_obs
            e = self.e_obs
            i = self.i_obs
            D = self.D_obs
            boxes = self.boxes_obs
        else:
            a = self.a_scl
            e = self.e_scl
            i = self.i_scl
            D = self.D_scl
            boxes = self.boxes_scl

        for j1 in xrange(0,self.n1):
            for j2 in xrange(0,self.n2):
                for j3 in xrange(0,self.n3):
                    for j4 in xrange(0,self.n4):
                        boxes[j1,j2,j3,j4] = 0

        for j in xrange(0,len(a)):
           j1 = int((a[j]-self.a1)/self.da)
           j2 = int((e[j]-self.e1)/self.de)
           j3 = int((i[j]-self.i1)/self.di)
           j4 = int((D[j]-self.D1)/self.dD)

           if (j1 >= 0) and (j1 < self.n1) and (j2 >= 0) and (j2 < self.n2) and (j3 >= 0) and (j3 < self.n3) and (j4 >= 0) and (j4 < self.n4) :
               if (j1,j2,j3,j4) in boxes.keys(): 
                   boxes[j1,j2,j3,j4] += 1
               else:
                   boxes[j1,j2,j3,j4] = 1

        #print "boxes = " + str(boxes)
        #print "len(boxes) = " + str(len(boxes))
        #print "min(boxes) = " + str(min(boxes.values()))
        #print "max(boxes) = " + str(max(boxes.values()))

    def save_boxes(self):
        """Save numbers of asteroids in boxes"""

        if len(self.boxes_scl) == 0:
            boxes = self.boxes_obs
            filename = "boxes_obs.dat"
        else:
            boxes = self.boxes_scl
            filename = "boxes_scl.dat"
            
        fout = file(filename, "w")

        for j1 in xrange(0,self.n1):
            for j2 in xrange(0,self.n2):
                for j3 in xrange(0,self.n3):
                    for j4 in xrange(0,self.n4):
                        a = self.a1 + j1*self.da
                        e = self.e1 + j2*self.de
                        i = self.i1 + j3*self.di
                        D = self.D1 + j4*self.dD
                        val = boxes[j1,j2,j3,j4]

                        fout.write("%.3f %.3f %.3f %.2f %5d\n" % (a, e, i, D, val))

            fout.write("\n")
        fout.close()

    def read_background(self):
        """Read observed background population"""

        i = 0
        self.D_bg = {}

        fin = open(self.file_background, "r")

        for line in fin:
            l = line.split()
            i += 1
            self.D_bg[i] = D_from_catalogue(l, self.pV)

        fin.close()
        print "number_of_background_asteroids = " + str(len(self.D_bg))

    def background_differential_sfd(self):    
        """Compute differential SFD of the background (including lists of bodies in the bins)"""

        self.dN_bg = []
        self.list_of_background = []
        for i in xrange(0,len(self.bins)):
            self.dN_bg.append(0)
            self.list_of_background.append([])

        for key in self.D_bg.keys():
            j = 1
            while self.D_bg[key] > self.bins[j]:
                j += 1
            self.dN_bg[j-1] += 1
            self.list_of_background[j-1].append(key)

        for i in xrange(0,len(self.bins)):
            self.dN_bg[i] = int(self.factor*self.dN_bg[i] + 0.5)

        print "factor = " + str(self.factor)
        print "dN_bg = " + str(self.dN_bg)

    def read_synthetic_diameters(self):
        """Read initial synthetic diameters (in km)"""
        
        i = 0
        self.D_syn = {}
        
        fin = open(self.file_diameters, "r")
        
        for line in fin:
            l = line.split()
            if len(l) > 1:
               i += 1
               self.D_syn[i] = float(l[0])*2./1000.  # R [m] -> D [km]
        
        fin.close()
        
        print "number_of_synthetic_diameters = " + str(len(self.D_syn))
        print "min(D_syn) = " + str(min(self.D_syn.values())) + " km"
        print "max(D_syn) = " + str(max(self.D_syn.values())) + " km"

    def read_synthetic_family(self):
        """Read synthetic family (sequentially for all times)"""

        a_syn = {}
        e_syn = {}
        i_syn = {}

        fin = open(self.file_synthetic, "r")
        fout = open("chi2.dat", "w")

        for line in fin:
            l = line.split()
            if len(l) > 0:
                id_ = int(l[0])
                t = float(l[1])
                a_syn[id_] = float(l[2])
                e_syn[id_] = float(l[3])
                i_syn[id_] = math.sin(float(l[4])/180.*math.pi)
            else:
                print "t = " + str(t) + " Myr ... number_of_synthetic_asteroids = " + str(len(a_syn))
                self.t = t
                self.a_syn = a_syn
                self.e_syn = e_syn
                self.i_syn = i_syn

                chi2, n = self.analyse()
                fout.write("%f %.6f %d\n" % (t, chi2, n))

                a_syn = {}
                e_syn = {}
                i_syn = {}

        fin.close()
        fout.close()

    def synthetic_differential_sfd(self):
        """Compute synthetic differential SFD (including lists of bodies in the bins)"""

        self.dN_syn = []
        self.list_of_synthetic = []
        for i in xrange(0,len(self.bins)):
            self.dN_syn.append(0)
            self.list_of_synthetic.append([])
 
        for key in self.a_syn.keys():
            j = 1
            while self.D_syn[key] > self.bins[j]:
                j += 1
            self.dN_syn[j-1] += 1
            self.list_of_synthetic[j-1].append(key)
 
        print "dN_syn = " + str(self.dN_syn)
        #print "list_of_synthetic = " + str(self.list_of_synthetic)

    def synthetic_background(self):
        """Create synthetic background in the given volume (a, e, sin I)"""

        a1_bg = 2.85
        a2_bg = 3.25
        e1_bg = self.e1
        e2_bg = self.e2
        i1_bg = self.i1
        i2_bg = self.i2
 
        self.a_scl = []
        self.e_scl = []
        self.i_scl = []
        self.D_scl = []

        for i in xrange(0,len(self.bins)):
            for j in xrange(0,self.dN_bg[i]):
                self.a_scl.append(random.uniform(a1_bg, a2_bg))
                self.e_scl.append(random.uniform(e1_bg, e2_bg))
                self.i_scl.append(random.uniform(i1_bg, i2_bg))
                key = random.choice(self.list_of_background[i])
                self.D_scl.append(self.D_bg[key])
 
        #print "D_scl = " + str(self.D_scl)

    def rescale_synthetic(self):
        """Re-scale synthetic SFD to the observed SFD (including synthetic orbital data, of course)"""

        for i in xrange(0,len(self.bins)):
            for j in xrange(0,self.dN_obs[i]-self.dN_bg[i]):
                key = random.choice(self.list_of_synthetic[i])
                self.a_scl.append(self.a_syn[key])
                self.e_scl.append(self.e_syn[key])
                self.i_scl.append(self.i_syn[key])
                self.D_scl.append(self.D_syn[key])

    def scaled_differential_sfd(self):
        """Compute scaled differential SFD (to check)"""

        self.dN_scl = []
        for i in xrange(0,len(self.bins)):
            self.dN_scl.append(0)

        for i in xrange(0,len(self.D_scl)):
            j = 1
            while self.D_scl[i] > self.bins[j]:
                j += 1
            self.dN_scl[j-1] += 1

        print "dN_scl = " + str(self.dN_scl)

    def save_scaled(self):
        """Save also scaled (synthetic) data"""

        fout = file("scl.dat", "w")
        for i in xrange(0,len(self.a_scl)):
            fout.write("%8.3f %4d %.5f %.5f %.5f %.5f\n" % (self.t, i, self.a_scl[i], self.e_scl[i], self.i_scl[i], self.D_scl[i]))
        fout.close()

    def create_images(self):
        """Create images and copy data files"""

        fout = open("t.plt", "w")
        fout.write("t = " + str(self.t))
        fout.close()
 
        subprocess.check_output("./ae_scl.plt", shell=True)
 
        str_ = "%04d" % (int(self.t+0.5))
        subprocess.check_output("cp ae_scl.png png/ae_scl_%s.png" % (str_), shell=True)

    def chi2(self):
        """Evaluate chi^2 metric"""
        chi2 = 0.
        n = 0
 
        for j1 in xrange(0,self.n1):
            for j2 in xrange(0,self.n2):
                for j3 in xrange(0,self.n3):
                    for j4 in xrange(0,self.n4):
                        a = self.a1 + j1*self.da
                        e = self.e1 + j2*self.de
                        i = self.i1 + j3*self.di
                        D = self.D1 + j4*self.dD
                        n_scl = self.boxes_scl[j1,j2,j3,j4]
                        n_obs = self.boxes_obs[j1,j2,j3,j4]
 
                        if n_scl > 0 and n_obs > 0:
                            sigma_scl = math.sqrt(n_scl)
                            sigma_obs = math.sqrt(n_obs)
 
                            chi2_ = (n_obs - n_scl)**2/(sigma_obs**2 + sigma_scl**2)
                            chi2 += chi2_
                            n += 1
                        else:
                            chi2_ = 0.

        return chi2, n

def main():
    blackbox = Blackbox()

if __name__ == "__main__":
    main()


